unit uRTFConverter;
// Must be at the top...
{$IF CompilerVersion >= 24.0}
  {$LEGACYIFEND ON}
{$IFEND}

interface
{$Region 'Changelog'}
{*********************************************************************************
**   Latest releases of uRTFConverter.pas are made available through the        **
**     distribution site at: http://www.audio-data.de/rtfconverter.html         **
**                                                                              **
**        See readme.txt for an introduction and documentation.                 **
**                                                                              **
**********************************************************************************
**    The contents of this file are used with permission, subject to            **
**    the Mozilla Public License Version 1.1 (the "License"); you may           **
**    not use this file except in compliance with the License. You may          **
**    obtain a copy of the License at                                           **
**    http:  www.mozilla.org/MPL/MPL-1.1.html                                   **
**                                                                              **
**    Software distributed under the License is distributed on an               **
**    "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or            **
**    implied. See the License for the specific language governing              **
**    rights and limitations under the License.                                 **
**                                                                              **
**    This unit based on the procedure RTF2Plain of the TICQLib package         **
**    http://sourceforge.net/projects/ticqlib/files/latest/download             **
**    Initial Developer of ICQWorks.RTF2Plain                                   **
**     Alex Demchenko(alex@ritlabs.com)                                         **
**                                                                              **
**    Contributor(s)                                                            **
**    (so)  Samuel Soldat     <samuel.soldat@audio-data.de>                     **
**                                                                              **
**********************************************************************************
**                                                                              **
** Created: 2012-April-22                                                       **
** (so) 2013-Feb-07: Eliminate range errors                                     **
** (so) 2016-Oct-06: adaption for XE4                                           **
** (so) 2017-Feb-02: adaption for Delphi 2007                                   **
**                                                                              **
*********************************************************************************}
{$EndRegion}

uses
  Windows, SysUtils, Classes;

type
  TRTF2PlainConverter = class
  private
    fPlaintext: String;
    fHyperlinks: array of AnsiString;
    fHashList: TList;
    function GetHyperlink(Idx: Integer): AnsiString;
    procedure AddHyperlink(const Target: AnsiString);
    function GetHyperlinkCount: Integer;
    function FindHashItem(CtrlId: Cardinal; CtrlWord: PAnsiChar): Pointer;
  public
    destructor Destroy; override;
    procedure ReadData(const Source: TBytes);
    property Plaintext: String read fPlaintext;
    property Hyperlink[Idx: Integer]: AnsiString read GetHyperlink;
    property HyperlinkCount: Integer read GetHyperlinkCount;
  end;

function IsRichText(const Source: TBytes): Boolean;

implementation

{$if RTLVersion >= 25} // since XE4
uses
  AnsiStrings;
{$ifend}

resourcestring
  sIllegalCharacterset = 'Illegal Characterset %d';

const
  cw_Line = 'line';
  cw_par = 'par';
  cw_bullet = 'bullet';
  cw_ldblquote = 'ldblquote';
  cw_rdblquote = 'rdblquote';
  cw_endash = 'endash';
  cw_emdash = 'emdash';
  cw_lquote = 'lquote';
  cw_rquote = 'rquote';
  cw_tab = 'tab';
  cw_cell = 'cell';
  cw_emspace = 'emspace';
  cw_enspace = 'enspace';
  cw_qmspace = 'qmspace';
  cw_unicode = 'u';
  cw_CPCount = 'uc';
  cw_font = 'f';
  cw_fcharset = 'fcharset';
  cw_fcodepage = 'cpg';
  cw_defaultcp = 'ansicpg';
  cw_pict = 'pict';
  cw_colortbl = 'colortbl';
  cw_fonttbl = 'fonttbl';
  cw_stylesheet = 'stylesheet';
  cw_info = 'info';
  cw_leveltext = 'leveltext';
  cw_levelnumbers = 'levelnumbers';
  cw_fldinst = 'fldinst';
  cw_listtable = 'listtable';
  cw_listoverridetable = 'listoverridetable';
  cw_nesttableprops = 'nesttableprops';

  tab_Char = #9;
  space_Char = ' ';
  lquote_Char = #145;
  rquote_Char = #146;
  ldblquote_Char = #147;
  rdblquote_Char = #148;
  bullet_Char = #149;
  endash_Char = #150;
  emdash_Char = #151;

type
  TCvtAction = (rtf_SpecialCharacter, rtf_UnicodeChar, rtf_SetCPCount, rtf_Ignore,
                rtf_GetText, rtf_Field, rtf_Fonttable, rtf_Font, rtf_Charset,
                rtf_Codepage, rtf_Defaultcp);
  TRtf_Control = record
    Hash: Cardinal;
    Word: PAnsiChar;
    Data: PAnsiChar;
    Action: TCvtAction;
  end;
  PRtf_Control = ^TRtf_Control;

const
  UnknownHashItem: TRtf_Control = (Action: rtf_Ignore);

const
  // Translation table charset->codepage
  CpTable: array [0..255] of Word =
//     0      1      2      3      4      5      6      7      8      9
    ( 1252,     0,    42, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
{50} 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 10000, 10001, 10003,
     10008, 10002, 65535, 10005, 10004, 10006, 10081, 10021, 10029, 10007,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
{100}65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,   932,   949,
      1361, 65535, 65535, 65535,   936, 65535,   950, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
{150}65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535,  1253,  1254,  1258, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535,  1255,  1256, 65535,
     65535, 65535, 65535, 65535, 65535, 65535,  1257, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
{200}65535, 65535, 65535, 65535,  1251, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535,   874, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,  1250, 65535,
     65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535,
{250}65535, 65535, 65535, 65535,   437,   850);

function ElfHashLowercase(var Hash: Cardinal; Ch: Byte): Boolean;
var
  g: Cardinal;
begin
  Result := (Ch>=Ord('a')) and (Ch<=Ord('z'));
  if not Result
  then begin
    Result := (Ch>=Ord('A')) and (Ch<=Ord('Z'));
    Inc(Ch, Ord('a') - Ord('A'));
  end;
  if Result
  then begin
    Hash := (Hash shl 4) + Ch;
    g := Hash and $f0000000;
    Hash := (Hash xor (g shr 24)) and not g;
  end;
end;

function CalcHash(s: PAnsiChar): Cardinal;
begin
  Result := 0;
  while ElfHashLowercase(Result, Ord(s^)) do
    inc(s);
end;

function CompareControls(Item1, Item2: Pointer): Integer;
begin
  Result := Integer(TRtf_Control(Item1^).Hash) - Integer(TRtf_Control(Item2^).Hash);
end;

function InitHash: TList;
  procedure Add(CtrlWord, CtrlData: PAnsiChar; CtrlAction: TCvtAction);
  var
    ctrl: ^TRtf_Control;
  begin
    New(ctrl);
    ctrl.Hash := CalcHash(CtrlWord);
    ctrl.Word := CtrlWord;
    ctrl.Data := CtrlData;
    ctrl.Action := CtrlAction;
    Result.Add(ctrl);
  end;
begin
  Result := TList.Create;
  Add(cw_Line, sLineBreak, rtf_SpecialCharacter);
  Add(cw_par, sLineBreak, rtf_SpecialCharacter);
  Add(cw_bullet, bullet_Char, rtf_SpecialCharacter);
  Add(cw_ldblquote, ldblquote_Char, rtf_SpecialCharacter);
  Add(cw_rdblquote, rdblquote_Char, rtf_SpecialCharacter);
  Add(cw_endash, endash_Char, rtf_SpecialCharacter);
  Add(cw_emdash, emdash_Char, rtf_SpecialCharacter);
  Add(cw_lquote, lquote_Char, rtf_SpecialCharacter);
  Add(cw_rquote, rquote_Char, rtf_SpecialCharacter);
  Add(cw_tab, tab_Char, rtf_SpecialCharacter);
  Add(cw_cell, space_Char, rtf_SpecialCharacter);
  Add(cw_emspace, space_Char, rtf_SpecialCharacter);
  Add(cw_enspace, space_Char, rtf_SpecialCharacter);
  Add(cw_qmspace, space_Char, rtf_SpecialCharacter);
  Add(cw_unicode, nil, rtf_UnicodeChar);
  Add(cw_CPCount, nil, rtf_SetCPCount);
  Add(cw_pict, nil, rtf_Ignore);
  Add(cw_colortbl, nil, rtf_Ignore);
  Add(cw_fonttbl, nil, rtf_Fonttable);
  Add(cw_font, nil, rtf_Font);
  Add(cw_fcharset, nil, rtf_Charset);
  Add(cw_fcodepage, nil, rtf_Codepage);
  Add(cw_defaultcp, nil, rtf_Defaultcp);
  Add(cw_stylesheet, nil, rtf_Ignore);
  Add(cw_info, nil, rtf_Ignore);
  Add(cw_leveltext, nil, rtf_Ignore);
  Add(cw_levelnumbers, nil, rtf_Ignore);
  Add(cw_listtable, nil, rtf_GetText);
  Add(cw_listoverridetable, nil, rtf_GetText);
  Add(cw_nesttableprops, nil, rtf_GetText);
  Add(cw_fldinst, nil, rtf_Field);
  Result.Sort(CompareControls);
end;

{ TRTF2PlainConverter }

procedure TRTF2PlainConverter.AddHyperlink(const Target: AnsiString);
var
  L: Integer;
begin
  L := Length(fHyperlinks);
  SetLength(fHyperlinks, L + 1);
  fHyperlinks[L] := Target;
end;

destructor TRTF2PlainConverter.Destroy;
var
  i: Integer;
begin
  if fHashList<>nil
  then begin
    for i := 0 to fHashList.Count - 1 do
      FreeMem(fHashList[i]);
    fHashList.Free;
  end;
  inherited;
end;

function TRTF2PlainConverter.FindHashItem(CtrlId: Cardinal; CtrlWord: PAnsiChar): Pointer;
var
  L, H, I, C: Integer;
begin
  L := 0;
  H := fHashList.Count - 1;
  Result := @UnknownHashItem;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := Integer(TRtf_Control(fHashList[I]^).Hash) - Integer(CtrlId);
    if C < 0
    then
      L := I + 1
    else begin
      H := I - 1;
      if (C = 0)
      then begin
        {$if RTLVersion >= 25}
        if (AnsiStrings.StrComp(TRtf_Control(fHashList[I]^).Word, CtrlWord)=0)
        {$else}
        if (SysUtils.StrComp(TRtf_Control(fHashList[I]^).Word, CtrlWord)=0)
        {$ifend}
        then
          Result := fHashList[I];
        break;
      end;
    end;
  end;
end;

function TRTF2PlainConverter.GetHyperlink(Idx: Integer): AnsiString;
begin
  Result := fHyperlinks[Idx];
end;

function TRTF2PlainConverter.GetHyperlinkCount: Integer;
begin
  Result := Length(fHyperlinks);
end;

procedure TRTF2PlainConverter.ReadData(const Source: TBytes);
{Convert RTF enabled text to plain.}
var
  Src: PByte;
  NChar: Integer;
  Codepage: Word;
  Codepages: array of Word;
  RtfChar: Byte;

  function NextChar: Byte;
  begin
    if NChar>0
    then begin
      Result := Src^;
      Dec(NChar);
      Inc(Src);
    end
    else
      Result := 0;
    RtfChar := Result;
  end;

  procedure SkipGroup;
  // Skip all characters in the current group. This is used for skip picture
  // data in a fast way.
  var
    BracesOpened: Integer;
    function ProcessChar(Ch: Byte): Boolean;
    begin
      Result := False;
      case Ch of
        0:
          Result := True;
        Ord('{'):
          Inc(BracesOpened);
        Ord('}'):
          begin
            Dec(BracesOpened);
            if BracesOpened<=0
            then
              Result := True;
          end;
        Ord('\'): //Next char is escaped
          begin
            Inc(Src);
            Dec(NChar);
          end;
      end;
    end;
  begin
    BracesOpened := 1;
    //Process current character
    if ProcessChar(RtfChar) or (NChar<=0)
    then
      exit;
    //Process subsequent characters
    while (NChar>0) and not ProcessChar(Src^) do
    begin
      Inc(Src);
      Dec(NChar);
    end;
    //Pop last character of the group (='}')
    NextChar;
  end;

  function CharFromHex: Byte;
  var
    Digit: Integer;
    i: Integer;
  begin
    Result := 0;
    for i:=0 to 1 do
    begin
      Result := Result * 16;
      Digit := NextChar - Ord('0');
      if Digit<0
      then
        exit;
      if (Digit>9)
      then begin
        Dec(Digit, Ord('A') - Ord('0'));
        if Digit<0
        then
          exit;
        if (Digit>5)
        then begin
          Dec(Digit, Ord('a') - Ord('A'));
          if (Digit<0) or (Digit>5)
          then
            exit;
        end;
        Inc(Result, 10);
      end;
      Inc(Result, Byte(Digit));
    end;
  end;

  function IsHyperlink: Boolean;
  const
    sHYPERLINK = 'HYPERLINK';
  var
    hl: PAnsiChar;
  begin
    hl := sHYPERLINK;
    while (hl^<>#0) and (RtfChar=Ord(hl^)) do
    begin
      Inc(hl);
      NextChar;
    end;
    Result := (hl^=#0);
  end;

  procedure GetControlInfo(var HashItem: PRtf_Control; var NumericValue: Integer);
  var
    CtrlId, CtrlPos: Cardinal;
    Control: array [0..33] of AnsiChar;
    NegativNumericValue: Boolean;
  begin
    CtrlId := 0;
    CtrlPos := 0;
    while ElfHashLowercase(CtrlId, RtfChar) and (CtrlPos<SizeOf(Control)-1)
    do begin
      Control[CtrlPos] := AnsiChar(RtfChar);
      Inc(CtrlPos);
      NextChar;
    end;
    Control[CtrlPos] := #0;
    HashItem := FindHashItem(CtrlId, Control);
    //End auf alphabetical part - follow a Number ?
    NumericValue := 0;
    NegativNumericValue := (RtfChar = Ord('-'));
    if NegativNumericValue
    then
      NextChar;
    while RtfChar in [Ord('0')..Ord('9')] do
    begin
      NumericValue := NumericValue * 10 + RtfChar - Ord('0');
      NextChar;
    end;
    if NegativNumericValue
    then
      NumericValue := - NumericValue;
    if RtfChar in [13, Ord(' ')]
    then //Delimiter is white space or CrLf
      while NextChar in [13, 10] do;
  end;

  procedure ProcessFontTable;
  // Read codepages from font table and move to end of the font table group
  var
    HashItem: PRtf_Control;
    NumericValue: Integer;
    FontNumber: Integer;
    BracesOpened: Integer;
  begin
    FontNumber := 0;
    BracesOpened := 1;
    repeat
      case RtfChar of
        0: break;
        Ord('{'):  // inner Group: {\f0 ..} {\f1 ..} ...
          begin
            Inc(BracesOpened);
            NextChar;
          end;
        Ord('}'):  // End of Group
          begin
            Dec(BracesOpened);
            while NextChar in [10, 13] do;
          end;
        Ord('\'): // Control found
          begin
            if NextChar in [Ord('a')..Ord('z')]
            then begin
              GetControlInfo(HashItem, NumericValue);
              case HashItem.Action of
                rtf_Font:
                  begin
                    if High(Codepages)<NumericValue
                    then
                      SetLength(Codepages, NumericValue + 1);
                    Codepages[NumericValue] := Codepage;
                    FontNumber := NumericValue;
                  end;
                rtf_Charset:
                  begin
                    if (NumericValue>=0) and (NumericValue<=High(CpTable)) and
                       (CpTable[NumericValue]<>65535)
                    then begin
                      Codepages[FontNumber] := CpTable[NumericValue];
                      if Codepages[FontNumber] = 0
                      then
                        Codepages[FontNumber] := GetACP;
                    end
                    else
                      raise Exception.CreateFmt(sIllegalCharacterset, [NumericValue]);
                  end;
                rtf_Codepage:
                  begin
                    Codepages[FontNumber] := NumericValue;
                    if Codepages[FontNumber] = 0
                    then
                      Codepages[FontNumber] := GetACP;
                  end;
              end; // Case
            end; // if
          end; // Control found
        else begin
          NextChar;
        end;
      end; //Case
    until BracesOpened<=0;
  end;

  function ProcessGroupRecursevly(AnsiCharSkipCount: Integer; FontNumber: Integer): string;
  var
    TextLen: Integer;
    Text: TBytes;
    procedure OutputSingleChar(const aChar: Byte);
    begin
      if TextLen>=Length(Text)
      then
        SetLength(Text, Length(Text) + 512);
      Text[TextLen] := aChar;
      Inc(TextLen);
      NextChar;
    end;
    procedure OutputStr(s: PAnsiChar);
    begin
      while s^<>#0 do
      begin
        if TextLen>=Length(Text)
        then
          SetLength(Text, Length(Text) + 512);
        Text[TextLen] := Byte(s^);
        Inc(TextLen);
        Inc(s);
      end;
    end;
    {$ifdef Unicode}
    procedure AppendText;
    var
      L, CharCount: Integer;
    begin
      if TextLen>0
      then begin
        //Encoding can not used because Codepage=42 (Symbol) is not supported
        CharCount := MultiByteToWideChar(Codepages[FontNumber], 0, Pointer(Text), TextLen, nil, 0);
        L := Length(Result);
        SetLength(Result, L + CharCount);
        MultiByteToWideChar(Codepages[FontNumber], 0, Pointer(Text), TextLen, @Result[L+1], CharCount);
        TextLen := 0;
      end;
    end;
    {$else}
    procedure AppendText;
    var
      L: Integer;
    begin
      if TextLen>0
      then begin
        L := Length(Result);
        SetLength(Result, L + TextLen);
        Move(Pointer(Text)^, Result[L+1], TextLen);
        TextLen := 0;
      end;
    end;
    {$endif}

  var // for ProcessGroupRecursevly
    Destination: Boolean;
    StarControl: Boolean;

    procedure ProcessControl;
    //process next control and move cursor to the first character after the control
    var
      HashItem: PRtf_Control;
      Link: AnsiString;
      NumericValue: Integer;
    begin
      NextChar;
      case RtfChar of
        Ord(''''): OutputSingleChar(CharFromHex); //Numeric AnsiChar
        Ord('~'):  OutputSingleChar(Ord(' '));
        Ord('-'):  NextChar;  //Optional hyphen - ignored
        Ord('_'):  OutputSingleChar(Ord('-'));
        Ord('\'):  OutputSingleChar(Ord('\'));
        Ord('{'):  OutputSingleChar(Ord('{'));
        Ord('|'):  OutputSingleChar(Ord('|'));  //Formula (not implemented)
        Ord(':'):  OutputSingleChar(Ord(':'));  //Specifies a subentry in an index entry (not implemented)
        Ord('}'):  OutputSingleChar(Ord('}'));
        Ord('*'):
          begin //Marks a destination whose text should be ignored if not understood by the RTF reader
            StarControl := True;
            NextChar;
          end;
        Ord('a')..Ord('z'):
          begin // control token
            GetControlInfo(HashItem, NumericValue);
            if (not Destination)
            then begin
              // first Control after '{' could be Destination
              Destination := True;
              case HashItem.Action of
                rtf_Ignore:
                  begin
                    if StarControl or (HashItem<>@UnknownHashItem)
                    then
                      SkipGroup;
                    exit;
                  end;
                rtf_Field:
                  begin
                    if IsHyperlink
                    then begin
                      repeat
                        NextChar;
                      until (RtfChar=Ord('"'));
                      NextChar;
                      Link := '';
                      while RtfChar<>Ord('"') do
                      begin
                        Link := Link + AnsiChar(RtfChar);
                        NextChar;
                      end;
                      NextChar;
                      AddHyperlink(Link);
                    end;
                    SkipGroup;
                    exit;
                  end;
                rtf_Fonttable:
                  begin
                    ProcessFontTable;
                    exit;
                  end;
              end;
            end;
            case HashItem.Action of
              rtf_SpecialCharacter:
                OutputStr(HashItem.Data); //Text := Text + HashItem.Data;
              {$ifdef Unicode}
              rtf_UnicodeChar:
                begin
                  // ToDo: If unicode enabled, I have to skip AnsiCharSkipCount chars
                  AppendText;
                  Result := Result + Char(NumericValue);
                end;
              {$else}
              rtf_UnicodeChar: ;
              {$endif}
              rtf_SetCPCount:
                AnsiCharSkipCount := NumericValue;
              rtf_Defaultcp:
                begin
                  if NumericValue=0
                  then
                    Codepage := GetACP
                  else
                    Codepage := NumericValue;
                  Codepages[0] := Codepage;
                end;
              rtf_Font:
                begin
                  //Change Font
                  if (NumericValue>=0) and (NumericValue<=High(Codepages))
                  then begin
                    if (FontNumber<>NumericValue) and
                       (Codepages[FontNumber]<>Codepages[NumericValue])
                    then begin
//                      outputdebugstring(PChar(Format('CP Wechsel %d->%d', [Codepages[FontNumber], Codepages[NumericValue]])));
                      AppendText;
                    end;
                    FontNumber := NumericValue;
                  end;
                  exit;
                end;
            end; //case
          end;
        else begin
          // Escaped Character
          OutputSingleChar(RtfChar);
        end;
      end; //Case
    end;
  begin // ProcessGroupRecursevly
    Result := '';
    TextLen := 0;
    Destination := False;
    StarControl := False;
    NextChar;
    while true do
    begin
      case RtfChar of
        0: Break;
        10, 13: NextChar; //Ignore Linebreak
        Ord('{'):  // inner Group
          begin
            AppendText;
            Result := Result + ProcessGroupRecursevly(AnsiCharSkipCount, FontNumber);
          end;
        Ord('}'):  // End of current Group
          begin
            while NextChar in [10, 13] do;
            Break;
          end;
        Ord('\'): // Control found
          ProcessControl;
        else begin
          //Plain Ansitext found
          OutputSingleChar(RtfChar);
        end;
      end; //Case
    end; // While
    AppendText;
  end; // ProcessGroupRecursevly
var
  temp: AnsiString;
begin
  SetLength(fHyperlinks, 0);
  NChar := Length(Source);
  if IsRichText(Source)
  then begin
    if fHashList=nil
    then
      fHashList := InitHash;
    SetLength(Codepages, 1);
    Codepages[0] := CPTable[0];
    Codepage := CPTable[0];
    Src := Pointer(Source);
    fPlaintext := ProcessGroupRecursevly(1, 0);
  end
  else
  if Source<>nil
  then begin
//    NChar := Length(Source); //StrLen(PAnsiChar(Pointer(aSource)));
    SetLength(temp, NChar);
    Move(Pointer(Source)^, Pointer(temp)^, NChar);
    fPlaintext := String(temp);
  end;
end;

function IsRichText(const Source: TBytes): Boolean;
const
  rtfHeader: array [0..5] of Byte = (Ord('{'), Ord('\'), Ord('r'), Ord('t'), Ord('f'), 0);
var
  p1, p2: PByte;
begin
  p1 := @rtfHeader;
  p2 := Pointer(Source);
  Result := Length(Source)>Length(rtfHeader);
  while Result and (p1^>0) do
  begin
    if p1^<>p2^
    then
      Result := False;
    Inc(p1); Inc(p2);
  end;
end;

end.
